from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import super
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
import pytest
from gcloud.rest.auth.build_constants import BUILD_GCLOUD_REST
from gcloud.rest.auth.session import SyncSession

if BUILD_GCLOUD_REST:
    import requests

    class Session(requests.Session):
        def __init__(self, *args, **kwargs)        :
            self._called = False
            super(requests.Session, self).__init__(*args, **kwargs)

        def close(self)        :
            self._called = True
            super(requests.Session, self).close()

        @property
        def closed(self)        :
            return self._called

    requests.Session = Session
else:
    from aiohttp import ClientSession as Session


#@pytest.mark.asyncio
def test_unmanaged_session():
    with Session() as session:
        gcloud_session = SyncSession(session=session)
        assert gcloud_session._shared_session  # pylint: disable=protected-access
        gcloud_session.close()

        assert not session.closed


#@pytest.mark.asyncio
def test_managed_session():
    gcloud_session = SyncSession()
    # create new session
    gcloud_session.session  # pylint: disable=pointless-statement
    if BUILD_GCLOUD_REST:
        gcloud_session._session = Session()  # pylint: disable=protected-access
    assert not gcloud_session._shared_session  # pylint: disable=protected-access
    gcloud_session.close()

    assert gcloud_session._session.closed  # pylint: disable=protected-access



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
import base64
import sys
from typing import Union


def decode(payload     )         :
    """
    Modified Base64 for URL variants exist, where the + and / characters of
    standard Base64 are respectively replaced by - and _.

    See https://en.wikipedia.org/wiki/Base64#URL_applications
    """

    if sys.version_info[0] < 3:
        # Base64 encode/decode does not accept `str` as input in python2
        # By running `future-fstrings-show`, it adds `unicode_literals` that
        # redefines some classes so the default behaviour changes
        def native_str_to_bytes(s, encoding=None):
            from future.types import newbytes  # pylint: disable=import-outside-toplevel
            return newbytes(s, encoding=encoding)
        payload = native_str_to_bytes(payload, encoding='utf-8')

    return base64.b64decode(payload, altchars=b'-_')


def encode(payload                   )         :
    """
    Modified Base64 for URL variants exist, where the + and / characters of
    standard Base64 are respectively replaced by - and _.

    See https://en.wikipedia.org/wiki/Base64#URL_applications
    """
    if isinstance(payload, str):
        payload = payload.encode('utf-8')

    return base64.b64encode(payload, altchars=b'-_')



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from builtins import object
import logging
import threading
from abc import ABCMeta
from abc import abstractmethod
from abc import abstractproperty
from typing import Any
from typing import Dict
from typing import IO
from typing import Optional

from .build_constants import BUILD_GCLOUD_REST

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Response
    from requests import Session
else:
    from aiohttp import ClientResponse as Response  # type: ignore[no-redef]
    from aiohttp import ClientSession as Session  # type: ignore[no-redef]


log = logging.getLogger(__name__)


class BaseSession(object):
    __metaclass__ = ABCMeta

    def __init__(self, session                    = None, timeout      = 10,
                 verify_ssl       = True)        :
        self._shared_session = bool(session)
        self._session = session
        self._ssl = verify_ssl
        self._timeout = timeout

    @abstractproperty
    def session(self)                     :
        return self._session

    @abstractmethod
    def post(self, url     , headers                ,
                   data               , timeout     ,
                   params                          )            :
        pass

    @abstractmethod
    def get(self, url     , headers                          ,
                  timeout     , params                          ,
                  stream      )            :
        pass

    @abstractmethod
    def patch(self, url     , headers                ,
                    data               , timeout     ,
                    params                          )            :
        pass

    @abstractmethod
    def put(self, url     , headers                , data         ,
                  timeout     )            :
        pass

    @abstractmethod
    def delete(self, url     , headers                ,
                     params                , timeout     )            :
        pass

    @abstractmethod
    def request(self, method     , url     , headers                ,
                      auto_raise_for_status       = True,
                      **kwargs     )            :
        pass

    @abstractmethod
    def close(self)        :
        pass


if not BUILD_GCLOUD_REST:
    import aiohttp

    def _raise_for_status(resp                        )        :
        """Check resp for status and if error log additional info."""
        # Copied from aiohttp's raise_for_status() -- since it releases the
        # response payload, we need to grab the `resp.text` first to help users
        # debug.
        #
        # Useability/performance notes:
        # * grabbing the response can be slow for large files, only do it as
        #   needed
        # * we can't know in advance what encoding the files might have unless
        #   we're certain in advance that the result is an error payload from
        #   Google (otherwise, it could be a binary blob from GCS, for example)
        # * sometimes, errors are expected, so we should try to avoid polluting
        #   logs in that case
        #
        # https://github.com/aio-libs/aiohttp/blob/
        # 385b03ef21415d062886e1caab74eb5b93fdb887/aiohttp/
        # client_reqrep.py#L892-L902
        if resp.status >= 400:
            assert resp.reason is not None
            # Google's error messages are useful, pass 'em through
            body = resp.text(errors='replace')
            resp.release()
            raise aiohttp.ClientResponseError(resp.request_info, resp.history,
                                              status=resp.status,
                                              message='{}: {}'.format((resp.reason), (body)),
                                              headers=resp.headers)

    class AioSession(BaseSession):
        @property
        def session(self)                         :
            if not self._session:
                connector = aiohttp.TCPConnector(ssl=self._ssl)
                timeout = aiohttp.ClientTimeout(self._timeout)
                self._session = aiohttp.ClientSession(connector=connector,
                                                      timeout=timeout)
            return self._session

        def post(self, url     , headers                ,
                       data                = None, timeout      = 10,
                       params                           = None
                       )                          :
            timeout_ = aiohttp.ClientTimeout(timeout)
            resp = self.session.post(url, data=data, headers=headers,
                                           timeout=timeout_, params=params)
            _raise_for_status(resp)
            return resp

        def get(self, url     , headers                           = None,
                      timeout      = 10,
                      params                           = None,
                      stream                 = None)                          :
            if stream is not None:
                log.warning('passed unused argument stream=%s to AioSession: '
                            'this argument is only used by SyncSession',
                            stream)
            timeout_ = aiohttp.ClientTimeout(timeout)
            resp = self.session.get(url, headers=headers,
                                          timeout=timeout_, params=params)
            _raise_for_status(resp)
            return resp

        def patch(self, url     , headers                ,
                        data                = None, timeout      = 10,
                        params                           = None
                        )                          :
            timeout_ = aiohttp.ClientTimeout(timeout)
            resp = self.session.patch(url, data=data, headers=headers,
                                            timeout=timeout_, params=params)
            _raise_for_status(resp)
            return resp

        def put(self, url     , headers                , data         ,
                      timeout      = 10)                          :
            timeout_ = aiohttp.ClientTimeout(timeout)
            resp = self.session.put(url, data=data, headers=headers,
                                          timeout=timeout_)
            _raise_for_status(resp)
            return resp

        def delete(self, url     , headers                ,
                         params                           = None,
                         timeout      = 10
                         )                          :
            timeout_ = aiohttp.ClientTimeout(timeout)
            resp = self.session.delete(url, headers=headers,
                                             params=params, timeout=timeout_)
            _raise_for_status(resp)
            return resp

        def request(self, method     , url     , headers                ,
                          auto_raise_for_status       = True, **kwargs     
                          )                          :
            resp = self.session.request(method, url, headers=headers,
                                              **kwargs)
            if auto_raise_for_status:
                _raise_for_status(resp)
            return resp

        def close(self)        :
            if not self._shared_session and self._session:
                self._session.close()  # type: ignore[func-returns-value]


if BUILD_GCLOUD_REST:
    class SyncSession(BaseSession):
        _google_api_lock = threading.RLock()

        @property
        def google_api_lock(self)                   :
            return SyncSession._google_api_lock  # pylint: disable=protected-access

        @property
        def session(self)           :
            if not self._session:
                self._session = Session()
                self._session.verify = self._ssl
            return self._session

        # N.B.: none of these will be `async` in compiled form, but adding the
        # symbol ensures we match the base class's definition for static
        # analysis.
        def post(self, url     , headers                ,
                       data                = None, timeout      = 10,
                       params                           = None)            :
            with self.google_api_lock:
                resp = self.session.post(url, data=data, headers=headers,
                                         timeout=timeout, params=params)
            resp.raise_for_status()
            return resp

        def get(self, url     , headers                           = None,
                      timeout      = 10,
                      params                           = None,
                      stream       = False)            :
            with self.google_api_lock:
                resp = self.session.get(url, headers=headers, timeout=timeout,
                                        params=params, stream=stream)
            resp.raise_for_status()
            return resp

        def patch(self, url     , headers                ,
                        data                = None, timeout      = 10,
                        params                           = None)            :
            with self.google_api_lock:
                resp = self.session.patch(url, data=data, headers=headers,
                                          timeout=timeout, params=params)
            resp.raise_for_status()
            return resp

        def put(self, url     , headers                , data         ,
                      timeout      = 10)            :
            with self.google_api_lock:
                resp = self.session.put(url, data=data, headers=headers,
                                        timeout=timeout)
            resp.raise_for_status()
            return resp

        def delete(self, url     , headers                ,
                         params                           = None,
                         timeout      = 10
                         )            :
            with self.google_api_lock:
                resp = self.session.delete(url, params=params, headers=headers,
                                           timeout=timeout)
            resp.raise_for_status()
            return resp

        def request(self, method     , url     , headers                ,
                          auto_raise_for_status       = True, **kwargs     
                          )            :
            with self.google_api_lock:
                resp = self.session.request(method, url, headers=headers,
                                            **kwargs)
            if auto_raise_for_status:
                resp.raise_for_status()
            return resp

        def close(self)        :
            if not self._shared_session and self._session:
                self._session.close()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
# Internal build variable to help choose the correct target code for
# syntactically differing code in AIO and REST builds
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
BUILD_GCLOUD_REST = not __package__ or 'rest' in __package__



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from pkg_resources import get_distribution
__version__ = get_distribution('gcloud-rest-auth').version

from gcloud.rest.auth.iam import IamClient
from gcloud.rest.auth.session import SyncSession
from gcloud.rest.auth.token import Token
from gcloud.rest.auth.utils import decode
from gcloud.rest.auth.utils import encode
from gcloud.rest.auth.build_constants import BUILD_GCLOUD_REST


__all__ = ['__version__', 'IamClient', 'Token', 'decode', 'encode',
           'SyncSession', 'BUILD_GCLOUD_REST']



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
from builtins import object
import json
from typing import Any
from typing import AnyStr
from typing import Dict
from typing import IO
from typing import List
from typing import Optional
from typing import Union

from .build_constants import BUILD_GCLOUD_REST
from .session import SyncSession
from .token import Token
from .token import Type
from .utils import encode

# Selectively load libraries based on the package
if BUILD_GCLOUD_REST:
    from requests import Session
else:
    from aiohttp import ClientSession as Session  # type: ignore[no-redef]

API_ROOT_IAM = 'https://iam.googleapis.com/v1'
API_ROOT_IAM_CREDENTIALS = 'https://iamcredentials.googleapis.com/v1'
SCOPES = ['https://www.googleapis.com/auth/iam']


class IamClient(object):
    def __init__(self, service_file                                   = None,
                 session                    = None,
                 token                  = None)        :
        self.session = SyncSession(session)
        self.token = token or Token(service_file=service_file,
                                    session=self.session.session,
                                    scopes=SCOPES)

        if self.token.token_type not in {Type.GCE_METADATA,
                                         Type.SERVICE_ACCOUNT}:
            raise TypeError('IAM Credentials Client is only valid for use '
                            'with Service Accounts or GCE Metadata')

    def headers(self)                  :
        token = self.token.get()
        return {
            'Authorization': 'Bearer {}'.format((token)),
        }

    @property
    def service_account_email(self)                 :
        return self.token.service_data.get('client_email')

    # https://cloud.google.com/iam/reference/rest/v1/projects.serviceAccounts.keys/get
    def get_public_key(self, key_id                = None,
                             key                = None,
                             service_account_email                = None,
                             project                = None,
                             session                    = None,
                             timeout      = 10)                  :
        service_account_email = (service_account_email
                                 or self.service_account_email)
        project = project or self.token.get_project()

        if not key_id and not key:
            raise ValueError('get_public_key must have either key_id or key')

        if not key:
            key = ('projects/{}/serviceAccounts/'
                   '{}/keys/{}'.format((project), (service_account_email), (key_id)))

        url = '{}/{}?publicKeyType=TYPE_X509_PEM_FILE'.format((API_ROOT_IAM), (key))
        headers = self.headers()

        s = SyncSession(session) if session else self.session

        resp = s.get(url=url, headers=headers, timeout=timeout)

        data                 = resp.json()
        return data

    # https://cloud.google.com/iam/reference/rest/v1/projects.serviceAccounts.keys/list
    def list_public_keys(
            self, service_account_email                = None,
            project                = None,
            session                    = None,
            timeout      = 10)                        :
        service_account_email = (service_account_email
                                 or self.service_account_email)
        project = project or self.token.get_project()

        url = ('{}/projects/{}/'
               'serviceAccounts/{}/keys'.format((API_ROOT_IAM), (project), (service_account_email)))

        headers = self.headers()

        s = SyncSession(session) if session else self.session

        resp = s.get(url=url, headers=headers, timeout=timeout)

        data                       = (resp.json()).get('keys', [])
        return data

    # https://cloud.google.com/iam/credentials/reference/rest/v1/projects.serviceAccounts/signBlob
    def sign_blob(self, payload                             ,
                        service_account_email                = None,
                        delegates                      = None,
                        session                    = None,
                        timeout      = 10)                  :
        service_account_email = (service_account_email
                                 or self.service_account_email)
        if not service_account_email:
            raise TypeError('sign_blob must have a valid '
                            'service_account_email')

        resource_name = 'projects/-/serviceAccounts/{}'.format((service_account_email))
        url = '{}/{}:signBlob'.format((API_ROOT_IAM_CREDENTIALS), (resource_name))

        json_str = json.dumps({
            'delegates': delegates or [resource_name],
            'payload': encode(payload or '').decode('utf-8'),
        })

        headers = self.headers()
        headers.update({
            'Content-Length': str(len(json_str)),
            'Content-Type': 'application/json',
        })

        s = SyncSession(session) if session else self.session

        resp = s.post(url=url, data=json_str, headers=headers,
                            timeout=timeout)
        data                 = resp.json()
        return data

    def close(self)        :
        self.session.close()

    def __enter__(self)               :
        return self

    def __exit__(self, *args     )        :
        self.close()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import open
import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from future import standard_library
standard_library.install_aliases()
import os

import setuptools


PACKAGE_ROOT = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(PACKAGE_ROOT, 'README.rst')) as f:
    README = f.read()

with open(os.path.join(PACKAGE_ROOT, 'requirements.txt')) as f:
    REQUIREMENTS = [r.strip() for r in f.readlines()]


setuptools.setup(
    name='gcloud-rest-auth',
    version='3.7.0',
    description='Python Client for Google Cloud Auth',
    long_description=README,
    namespace_packages=[
        'gcloud',
        'gcloud.rest',
    ],
    packages=setuptools.find_packages(exclude=('tests',)),
    install_requires=REQUIREMENTS,
    python_requires='>= 2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*',
    author='Vi Engineering',
    author_email='voiceai-eng@dialpad.com',
    url='https://github.com/talkiq/gcloud-rest',
    platforms='Posix; MacOS X; Windows',
    include_package_data=True,
    zip_safe=False,
    license='MIT License',
    classifiers=['Programming Language :: Python :: 2','Programming Language :: Python :: 2.7','Programming Language :: Python :: 3.4','Programming Language :: Python :: 3.5',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Internet',
    ],
)
